import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.svetovid.io.SvetovidReader;
import org.svetovid.io.SvetovidWriter;

@interface DatoNaVezbama {}

// Ako bude potrebna, ova klasa ce biti data na testu
@DatoNaVezbama
public class StabloIOPretty<T> {

    protected static final String EMPTY_SYMBOL = " ";
    protected static final String RIGHT_SYMBOL = "/";
    protected static final String VERTICAL_SYMBOL = "|";
    protected static final String LEFT_SYMBOL = "\\";
    protected static final String HORIZONTAL_SYMBOL = "-";

    protected Konverter<T> konverter;
    protected String nullSymbol;
    protected int separation;
    protected int length;

    public StabloIOPretty(Konverter<T> konverter) {
        this(konverter, null, 1, 7);
    }

    public StabloIOPretty(Konverter<T> konverter, String nullSymbol, int separation, int length) {
    	this.konverter = konverter;
        this.nullSymbol = nullSymbol;
        this.separation = separation;
        this.length = length;
    }

    public Konverter<T> getKonverter() {
		return konverter;
	}

	public void setKonverter(Konverter<T> konverter) {
		this.konverter = konverter;
	}

	public String getNullSymbol() {
        return nullSymbol;
    }

    public void setNullSymbol(String nullSymbol) {
        this.nullSymbol = nullSymbol;
    }

    public int getSeparation() {
        return separation;
    }

    public void setSeparation(int separation) {
        this.separation = separation;
    }

    public int getLength() {
        return length;
    }

    public void setLength(int length) {
        if (length < 3) {
            throw new IllegalArgumentException("length");
        }
        this.length = length;
    }

    public BinarnoStablo<T> readStablo(SvetovidReader in) {
        return parseStablo(in, konverter, nullSymbol, length);
    }

    protected BinarnoStablo<T> parseStablo(SvetovidReader in, Konverter<T> konverter, String nullSymbol, int length) {
        List<BinarnoStablo<T>> elements = new ArrayList<>();
        List<Integer> levels = new ArrayList<>();
        Pattern levelPattern = Pattern.compile("[\\Q" + LEFT_SYMBOL + HORIZONTAL_SYMBOL + RIGHT_SYMBOL + "\\E]");
        String line = in.readLine();
        while ((line != null) && !line.isEmpty()) {
            Matcher matcher = levelPattern.matcher(line);
            int level = -1;
            if (matcher.find()) {
                level = matcher.start();
            }
            if (level != -1 && (nullSymbol == null || !line.endsWith(nullSymbol))) {
            	BinarnoStablo<T> stablo = parseStablo(konverter, line);
                elements.add(stablo);
                levels.add(level);
            }
            line = in.readLine();
        }
        BinarnoStablo<T> stablo = formStablo(0, elements.size(), levels, elements);
        return stablo;
    }

    private BinarnoStablo<T> parseStablo(Konverter<T> konverter, String line) {
        String vrednost;
        int beginIndex = line.indexOf('(');
        int endIndex = line.indexOf(')');
        if ((beginIndex != -1) && (endIndex != -1) && (beginIndex < endIndex)) {
            vrednost = line.substring(endIndex + 2);
        } else {
            throw new NumberFormatException(line);
        }
        T element = konverter.fromString(vrednost);
        BinarnoStablo<T> stablo = new BinarnoStablo<>(element);
        return stablo;
    }

    private BinarnoStablo<T> formStablo(int beginIndex, int endIndex, List<Integer> levels, List<BinarnoStablo<T>> elements) {
        if (beginIndex >= endIndex) {
            return null;
        }
        int minIndex = beginIndex;
        int minLevel = levels.get(minIndex);
        for (int i = beginIndex + 1; i < endIndex; i++) {
            int level = levels.get(i);
            if (level < minLevel) {
                minLevel = level;
                minIndex = i;
            }
        }
        BinarnoStablo<T> stablo = elements.get(minIndex);
        BinarnoStablo<T> levo = formStablo(minIndex + 1, endIndex, levels, elements);
        BinarnoStablo<T> desno = formStablo(beginIndex, minIndex, levels, elements);
        stablo.setLevoPodstablo(levo);
        stablo.setDesnoPodstablo(desno);
        return stablo;
    }

    public void printStablo(SvetovidWriter out, BinarnoStablo<T> stablo) {
        StringBuilder builder = new StringBuilder();
        appendTree(builder, stablo, konverter, nullSymbol, separation, length);
        out.print(builder.toString());
    }

    protected void appendTree(StringBuilder builder, BinarnoStablo<T> stablo, Konverter<T> konverter, String nullSymbol, int separation, int length) {
        String[] buildingBlocks = generateBuildingBlocks(length);
        appendRight(builder, stablo, konverter, nullSymbol, separation, buildingBlocks, true, buildingBlocks[5]);
        appendNode(builder, stablo, konverter, nullSymbol != null ? nullSymbol : "|", buildingBlocks[4]);
        appendLeft(builder, stablo, konverter, nullSymbol, separation, buildingBlocks, false, buildingBlocks[5]);
    }

    protected void appendNode(StringBuilder builder, BinarnoStablo<T> stablo, Konverter<T> konverter, String nullSymbol, String prefix) {
        builder.append(prefix);
        if (stablo == null) {
            builder.append(nullSymbol);
        } else {
            builder.append("(o) ");
            String vrednost = konverter.toString(stablo.getElement());
            builder.append(vrednost);
        }
        builder.append("\n");
    }

    protected void appendRight(StringBuilder builder, BinarnoStablo<T> stablo, Konverter<T> konverter, String nullSymbol, int separation, String[] buildingBlocks, boolean isRight, String prefix) {
        if (stablo == null) {
            return;
        }
        if ((nullSymbol != null) || (stablo.getDesnoPodstablo() != null)) {
            appendSubtree(builder, stablo.getDesnoPodstablo(), konverter, nullSymbol, separation, buildingBlocks, true, prefix);
            for (int i = 0; i < separation; i++) {
                appendEmpty(builder, buildingBlocks, prefix);
            }
        }
    }

    protected void appendLeft(StringBuilder builder, BinarnoStablo<T> stablo, Konverter<T> konverter, String nullSymbol, int separation, String[] buildingBlocks, boolean isRight, String prefix) {
        if (stablo == null) {
            return;
        }
        if ((nullSymbol != null) || (stablo.getLevoPodstablo() != null)) {
            for (int i = 0; i < separation; i++) {
                appendEmpty(builder, buildingBlocks, prefix);
            }
            appendSubtree(builder, stablo.getLevoPodstablo(), konverter, nullSymbol, separation, buildingBlocks, false, prefix);
        }
    }

    protected void appendEmpty(StringBuilder builder, String[] buildingBlocks, String prefix) {
        builder.append(prefix);
        builder.append(buildingBlocks[2]);
        builder.append("\n");
    }

    protected void appendSubtree(StringBuilder builder, BinarnoStablo<T> stablo, Konverter<T> konverter, String nullSymbol, int separation, String[] buildingBlocks, boolean isRight, String prefix) {
        String mojPrefix = prefix;
        if (isRight == true) {
            mojPrefix = mojPrefix + buildingBlocks[1];
        }
        if (isRight == false) {
            mojPrefix = mojPrefix + buildingBlocks[3];
        }
        String noviPrefix = prefix + (!isRight ? buildingBlocks[2] : buildingBlocks[0]);
        appendRight(builder, stablo, konverter, nullSymbol, separation, buildingBlocks, isRight, noviPrefix);
        appendNode(builder, stablo, konverter, nullSymbol, mojPrefix);
        noviPrefix = prefix + (isRight ? buildingBlocks[2] : buildingBlocks[0]);
        appendLeft(builder, stablo, konverter, nullSymbol, separation, buildingBlocks, isRight, noviPrefix);
    }

    private String[] generateBuildingBlocks(int length) {
        String[] blocks = new String[6];
        blocks[0] = generateBlock(EMPTY_SYMBOL, EMPTY_SYMBOL, EMPTY_SYMBOL, length - 2);
        blocks[1] = generateBlock(EMPTY_SYMBOL, RIGHT_SYMBOL, HORIZONTAL_SYMBOL, length - 2);
        blocks[2] = generateBlock(EMPTY_SYMBOL, VERTICAL_SYMBOL, EMPTY_SYMBOL, length - 2);
        blocks[3] = generateBlock(EMPTY_SYMBOL, LEFT_SYMBOL, HORIZONTAL_SYMBOL, length - 2);
        blocks[4] = HORIZONTAL_SYMBOL;
        blocks[5] = EMPTY_SYMBOL;
        return blocks;
    }

    protected String generateBlock(String emptySymbol, String startSymbol, String repeatSymbol, int repeatCount) {
        StringBuilder builder = new StringBuilder();
        builder.append(emptySymbol);
        builder.append(startSymbol);
        for (int i = 0; i < repeatCount; i++) {
            builder.append(repeatSymbol);
        }
        return builder.toString();
    }
}
